"""
Interface classes for state files.

"""
from datreant.core.backends.statefiles import TreantFile


class SimFile(TreantFile):

    def _init_state(self):
        super(SimFile, self)._init_state()
        self._state['mdsynthesis'] = dict()



"""
Functions and classes for finding Treants in the filesystem.

"""
import os
import glob

from datreant.core import Treant, Group


class Universehound(object):
    """Locator for Universe files.

    This object is used by Sims to find their Universe files, even when
    they go missing.

    """
    def __init__(self, caller):
        """Generate a Universehound to track down Universe files.

        :Arguments:
            *caller*
                object that summoned the Univershound

        """
        self.caller = caller
        self.ufiles = list()

    def fetch(self):
        """Find the Universe files.

        For finding Universe files, the Foxhound begins by looking for
        the files with the same basename among the paths it was given.
        If the files can't be found in those places, it stops looking.

        Raises :exc:`IOError` if a file could not be found.

        :Returns:
            *results*
                dictionary giving 'top' and 'traj' as keys, with
                lists of topology and trajectory file(s) as
                values, respectively; ``None`` as a member of
                a list indicates that a particular file could not
                be found

        """
        # search last-known locations
        results = self._check_basedirs()

        if (results['top'] is None) or (None in results['traj']):
            raise IOError("At least one file for"
                          " universe could not"
                          " be found from stored absolute and relative"
                          " paths. Set correct paths manually.")

        # TODO: include hash check; will require stored hashes for each
        # universe file

        return results

    def _check_basedirs(self):
        """Check last-known locations for Universe files.

        :Returns:
            *results*
                dictionary giving 'top' and 'traj' as keys, with
                lists of topology and trajectory file(s) as
                values, respectively; ``None`` as a member of
                a list indicates that a particular file could not
                be found
        """
        top = dict()
        traj = dict()
        for pathtype in ('abs', 'rel'):
            top[pathtype], traj[pathtype] = self.caller._define(
                    pathtype=pathtype)
        outpaths = dict()

        def check(paths):

            out = {'abs': None,
                   'rel': None}

            chosen = None

            # check absolute path first
            if os.path.exists(paths['abs']):
                out['abs'] = paths['abs']

            candidate = os.path.join(self.caller._treant.abspath, paths['rel'])
            if os.path.exists(candidate):
                out['rel'] = candidate

            # if both abs and rel exist, check that they point to the same
            # file. if so, accept; if not, throw exception
            if ((out['abs'] and out['rel']) and not
                    os.path.samefile(out['abs'], out['rel'])):
                raise IOError("Absolute and relative paths for file '{}'"
                              " point to different files.".format(out['abs']))
            elif out['rel']:
                # otherwise, accept rel
                chosen = os.path.abspath(out['rel'])
            elif out['abs']:
                # if rel path gives no file, accept abs
                chosen = out['abs']
            else:
                # if none of the paths resolve, raise exception
                raise IOError(
                        "File '{}' not found for universe".format(out['abs']))

            return chosen

        outpaths['top'] = check(top)

        outpaths['traj'] = [None]*len(traj['abs'])
        for i, entry in enumerate(traj['abs']):
            outpaths['traj'][i] = check({'abs': traj['abs'][i],
                                         'rel': traj['rel'][i]})

        return outpaths



"""
Limb attachments for :mod:`mdsynthesis` classes.

"""
from datreant.core.treants import Group
from datreant.core.collections import Bundle, View
from datreant.data.limbs import Data
from datreant.data.agglimbs import AggData

from .treants import Sim

Sim._attach_limb_class(Data)
Group._attach_limb_class(Data)
Bundle._attach_agglimb_class(AggData)
View._attach_aggtreelimb_class(AggData)



"""Interface tests for Treants.

"""

import mdsynthesis as mds
import pandas as pd
import numpy as np
import pytest
import os
import shutil
import py
from pkg_resources import parse_version

from datreant.core.tests.test_treants import TestTreant

import MDAnalysis as mda
from MDAnalysisTests.datafiles import PDB, GRO, XTC, PSF


class TestSim(TestTreant):
    """Test Sim-specific features"""
    treantname = 'testsim'
    treanttype = 'Sim'

    @pytest.fixture
    def treantclass(self):
        return mds.Sim

    @pytest.fixture
    def treant(self, tmpdir):
        with tmpdir.as_cwd():
            s = mds.Sim(TestSim.treantname)
        return s

    class TestUniverse:
        """Test universe functionality"""

        def test_add_universe(self, treant):
            """Test adding a new unverse definition"""
            treant.universedef.topology = GRO
            treant.universedef.trajectory = XTC

            assert isinstance(treant.universe, mda.Universe)

            assert treant.universe.filename == GRO
            assert treant.universe.trajectory.filename == XTC

        def test_set_universe(self, treant):
            """Test setting the universe with a Universe object"""

            u = mda.Universe(GRO, XTC)

            # set the universe directly
            treant.universe = u

            assert isinstance(treant.universe, mda.Universe)
            assert treant.universedef.topology == GRO
            assert treant.universedef.trajectory == XTC

        def test_set_universe_only_topology(self, treant):
            """Test setting the Universe to a topology-only universe"""
            u = mda.Universe(PSF)

            treant.universe = u

            assert treant.universe.filename == PSF
            assert treant.universedef.topology == PSF
            assert treant.universedef.trajectory is None

        def test_set_universe_chainreader(self, treant):
            """Test setting the Universe to multiple file trajectory (chain)"""
            u = mda.Universe(GRO, (XTC, XTC))

            treant.universe = u

            assert treant.universedef.topology == GRO
            assert treant.universedef.trajectory == (XTC, XTC)

        @pytest.mark.skipif((parse_version(mda.__version__) <
                            parse_version('0.15.0')),
                            reason="requires MDAnalysis >= 0.15.0")
        def test_set_universe_with_kwargs(self, treant):
            """Universe should preserve its kwargs, if possible.

            Test that setting a Universe for a Sim also gets its kwargs
            preserved, that an exception is raised for unserializable kwargs,
            and that a proper warning is geven when the Sim can't get them from
            the Universe in the first place.

            """
            # try out a basic, but nonexistent, kwarg
            u = mda.Universe(PDB, XTC, something_fake=True)

            treant.universe = u
            assert treant.universedef.kwargs['something_fake'] is True

            # try a kwarg with a value that won't be serializable
            u2 = mda.Universe(PDB, XTC,
                              something_else=mda.topology.GROParser.GROParser)

            with pytest.raises(ValueError):
                treant.universe = u2

            # check that we get a warning if a Universe didn't store its kwargs
            u3 = mda.Universe(PDB, XTC, something_fake=True)
            del u3._kwargs
            with pytest.warns(UserWarning):
                treant.universe = u3

        def test_add_univese_typeerror(self, treant):
            """Test checking of what is passed to setter"""
            with pytest.raises(TypeError):
                treant.universe = 72

        def test_remove_universe(self, treant):
            """Test universe removal"""
            treant.universedef.topology = GRO
            treant.universedef.trajectory = XTC

            assert isinstance(treant.universe, mda.Universe)

            treant.universedef.topology = None
            assert treant.universedef.topology is None
            assert treant.universe is None

            # trajectory definition should still be there
            assert treant.universedef.trajectory

            # this should give us a universe again
            treant.universedef.topology = PDB

            assert isinstance(treant.universe, mda.Universe)

            # removing trajectories should keep universe, but with PDB as
            # coordinates
            treant.universedef.trajectory = None

            assert isinstance(treant.universe, mda.Universe)
            assert treant.universe.trajectory.n_frames == 1

        def test_set_resnums(self, treant):
            """Test that we can add resnums to a universe."""
            treant.universedef.topology = GRO
            treant.universedef.trajectory = XTC

            protein = treant.universe.select_atoms('protein')
            resids = protein.residues.resids
            protein.residues.set_resnum(resids + 3)

            treant.universedef._set_resnums(treant.universe.residues.resnums)

            treant.universedef.reload()

            protein = treant.universe.select_atoms('protein')
            assert (resids + 3 == protein.residues.resnums).all()

            # test resetting of resnums
            protein.residues.set_resnum(resids + 6)

            assert (protein.residues.resnums == resids + 6).all()
            treant.universedef._set_resnums(treant.universe.residues.resnums)

            treant.universedef.reload()

            protein = treant.universe.select_atoms('protein')
            assert (resids + 6 == protein.residues.resnums).all()

    class TestSelections:
        """Test stored atomselections functionality"""
        @pytest.fixture
        def treant(self, tmpdir):
            with tmpdir.as_cwd():
                s = mds.Sim(TestSim.treantname)

                s.universedef.topology = GRO
                s.universedef.trajectory = XTC
            return s

        def test_add_selection(self, treant):
            """Test adding new selection definitions"""

            treant.atomselections.add('CA', 'protein and name CA')
            treant.atomselections.add('someres', 'resid 12')

            CA = treant.universe.select_atoms('protein and name CA')
            someres = treant.universe.select_atoms('resid 12')

            assert (CA.indices ==
                    treant.atomselections.create('CA').indices).all()
            assert (someres.indices ==
                    treant.atomselections.create('someres').indices).all()

        def test_remove_selection(self, treant):
            """Test universe removal"""

            treant.atomselections.add('CA', 'protein and name CA')
            treant.atomselections.add('someres', 'resid 12')

            assert 'CA' in treant.atomselections
            assert 'someres' in treant.atomselections

            treant.atomselections.remove('CA')

            assert 'CA' not in treant.atomselections
            assert 'someres' in treant.atomselections

            del treant.atomselections['someres']
            treant.atomselections.add('moreres', 'resid 12:20')

            assert 'CA' not in treant.atomselections
            assert 'someres' not in treant.atomselections
            assert 'moreres' in treant.atomselections

            with pytest.raises(KeyError):
                treant.atomselections.remove('someres')

            with pytest.raises(KeyError):
                del treant.atomselections['CA']

        def test_selection_keys(self, treant):
            treant.atomselections.add('CA', 'protein and name CA')
            treant.atomselections.add('someres', 'resid 12')

            assert set(('CA', 'someres')) == set(treant.atomselections.keys())

        def test_selection_get(self, treant):
            CA = 'protein and name CA'
            treant.atomselections.add('CA', CA)

            assert treant.atomselections.get('CA') == CA
            assert treant.atomselections['CA'] == CA

        def test_selection_get_not_present(self, treant):
            with pytest.raises(KeyError):
                treant.atomselections['CA']

        def test_add_atomselections_multiple_strings_via_add(self, treant):
            """Add a selection that has multiple selection strings"""
            treant.atomselections.add('funky town', 'name N', 'name CA')
            assert 'funky town' in treant.atomselections

            ref = treant.universe.select_atoms('name N', 'name CA')
            sel = treant.atomselections.create('funky town')
            assert (ref.indices == sel.indices).all()

        def test_add_atomselections_multiple_strings_via_setitem(self, treant):
            """Add a selection that has multiple selection strings"""
            treant.atomselections['funky town 2'] = 'name N', 'name CA'
            assert 'funky town 2' in treant.atomselections

            ref = treant.universe.select_atoms('name N', 'name CA')
            sel = treant.atomselections.create('funky town 2')
            assert (ref.indices == sel.indices).all()

        def test_add_selection_as_indices_via_add(self, treant):
            """Make an arbitrary AtomGroup then save selection as AG"""
            ag = treant.universe.atoms[:10:2]

            treant.atomselections.add('ag sel', ag.indices)
            assert 'ag sel' in treant.atomselections

            ag2 = treant.atomselections.create('ag sel')
            assert (ag.indices == ag2.indices).all()

        def test_add_selection_as_indices_via_setitem(self, treant):
            """Make an arbitrary AtomGroup then save selection as AG"""
            ag = treant.universe.atoms[25:50:3]

            treant.atomselections['ag sel 2'] = ag.indices
            assert 'ag sel 2' in treant.atomselections

            ag2 = treant.atomselections.create('ag sel 2')
            assert (ag.indices == ag2.indices).all()

        def test_add_selection_as_mix_via_setitem(self, treant):
            """Save a selection as a mixture of atom indices and strings"""
            ag = treant.universe.atoms[25:50:3]

            treant.atomselections['ag sel 2'] = ('protein and name CA',
                                                 ag.indices)

            assert 'ag sel 2' in treant.atomselections

            ag2 = treant.atomselections.create('ag sel 2')

            ag3 = treant.universe.select_atoms('protein and name CA') + ag
            assert (ag2.indices == ag3.indices).all()

        def test_add_selection_as_mix_via_add(self, treant):
            """Save a selection as a mixture of atom indices and strings"""
            ag = treant.universe.atoms[25:50:3]

            treant.atomselections.add('ag sel 2', 'protein and name CA',
                                      ag.indices)

            assert 'ag sel 2' in treant.atomselections

            ag2 = treant.atomselections.create('ag sel 2')

            ag3 = treant.universe.select_atoms('protein and name CA') + ag
            assert (ag2.indices == ag3.indices).all()


class TestReadOnly:
    """Test Sim functionality when read-only"""

    GRO = 'md.gro'
    XTC = 'md.xtc'

    @pytest.fixture
    def sim(self, tmpdir, request):
        with tmpdir.as_cwd():
            c = mds.Sim('testsim')

            # copy universe files to within the Sim's tree
            sub = py.path.local(c.abspath).mkdir('sub')
            GRO_t = sub.join(self.GRO)
            XTC_t = sub.join(self.XTC)
            py.path.local(GRO).copy(GRO_t)
            py.path.local(XTC).copy(XTC_t)

            c.universedef.topology = GRO_t.strpath
            c.universedef.trajectory = XTC_t.strpath

            py.path.local(c.abspath).chmod(0550, rec=True)

        def fin():
            py.path.local(c.abspath).chmod(0770, rec=True)

        request.addfinalizer(fin)

        return c

    def test_sim_universe_access(self, sim):
        """Test that Sim can access Universe when read-only.
        """
        assert isinstance(sim.universe, mda.Universe)

    def test_sim_moved_universe_access(self, sim, tmpdir):
        """Test that Sim can access Universe when read-only, especially
        when universe files have moved with it (stale paths).
        """
        py.path.local(sim.abspath).chmod(0770, rec=True)
        sim.location = tmpdir.mkdir('test').strpath
        py.path.local(sim.abspath).chmod(0550, rec=True)

        assert isinstance(sim.universe, mda.Universe)

        py.path.local(sim.abspath).chmod(0770, rec=True)



"""Tests for filesystem elements, including Foxhound and related objects.

"""

import mdsynthesis as mds
import pytest
import py.path

import MDAnalysis
from MDAnalysisTests.datafiles import GRO, XTC


class TestUniversehound:
    """Test Universehound functionality"""
    GRO = 'md.gro'
    XTC = 'md.xtc'

    @pytest.fixture
    def sim_external(self, tmpdir):
        """Create a Sim object with an externally-stored set of
        universe files.
        """
        with tmpdir.as_cwd():
            s = mds.Sim('testsim')
            sub = tmpdir.mkdir('sub')
            GRO_t = sub.join(self.GRO)
            XTC_t = sub.join(self.XTC)
            py.path.local(GRO).copy(GRO_t)
            py.path.local(XTC).copy(XTC_t)

            s.universedef.topology = GRO_t.strpath
            s.universedef.trajectory = XTC_t.strpath
        return s

    @pytest.fixture
    def sim_internal(self, tmpdir):
        """Create a Sim object with an internally-stored set of
        universe files.
        """
        with tmpdir.as_cwd():
            s = mds.Sim('testsim')
            sub = py.path.local(s.abspath).mkdir('sub')
            GRO_t = sub.join(self.GRO)
            XTC_t = sub.join(self.XTC)
            py.path.local(GRO).copy(GRO_t)
            py.path.local(XTC).copy(XTC_t)

            s.universedef.topology = GRO_t.strpath
            s.universedef.trajectory = XTC_t.strpath
        return s

    def test_move_Sim_external(self, sim_external, tmpdir):
        """Test that when we move the Sim object, it can still
        find its externally-stored set of universe files.
        """
        sim_external.location = tmpdir.mkdir('test').strpath
        assert isinstance(sim_external.universe, MDAnalysis.Universe)

        sim = sim_external
        assert sim.universe.filename == sim.universedef.topology
        assert sim.universe.trajectory.filename == sim.universedef.trajectory

    def test_move_Sim_internal(self, sim_internal, tmpdir):
        sim_internal.location = tmpdir.mkdir('test').strpath
        assert isinstance(sim_internal.universe, MDAnalysis.Universe)

        sim = sim_internal
        assert sim.universe.filename == sim.universedef.topology
        assert sim.universe.trajectory.filename == sim.universedef.trajectory

    def test_move_Sim_internal_copy(self, sim_internal, tmpdir):
        """We move the whole Sim, including its internal trajectory files, and
        we put a copy of one trajectory file where the internal ones used to
        be.
        """
        grofile = tmpdir.join(sim_internal.name, 'sub', self.GRO)
        assert sim_internal.universedef.topology == grofile.strpath

        sim_internal.location = tmpdir.mkdir('test').strpath
        tmpdir.join('test', sim_internal.name, 'sub', self.GRO).copy(
                tmpdir.mkdir(sim_internal.name).mkdir('sub'))

        with pytest.raises(IOError):
            sim_internal.universe



"""
Limbs are user interfaces for accessing stored data, as well as querying
the state of an object (data loaded, universe attached, etc.). They are also
used to aggregate the functionality of higher level objects (such as Sim) in
ways that are user-friendly.

In short, an Limb is designed to be user friendly on its own, but are
often used as components of a Treant.

"""
import os
from six import string_types
import numpy as np
from numpy.lib.utils import deprecate
import warnings

from datreant.core import Leaf
from datreant.core.limbs import Limb
from MDAnalysis import Universe
from MDAnalysis.core.AtomGroup import AtomGroup

from .filesystem import Universehound


class UniverseDefinition(Limb):
    """The defined universe of the Sim.

    Universes are interfaces to raw simulation data, with stored selections for
    this universe directly available via ``Sim.atomselections``.

    """
    _name = 'universedef'
    _filepaths = ['abs', 'rel']

    def __init__(self, treant):
        super(UniverseDefinition, self).__init__(treant)

        subitems = {'topology': dict,
                    'trajectory': list,
                    'kwargs': dict}

        # init state if for udef not already there;
        # if read-only, check that it is there,
        # and raise exception if it is not
        try:
            with self._treant._write:
                mdsdict = self._treant._state['mdsynthesis']
                try:
                    mdsdict[self._name]
                except KeyError:
                    mdsdict[self._name] = dict()

                for key, value in subitems.items():
                    try:
                        mdsdict[self._name][key]
                    except KeyError:
                        mdsdict[self._name][key] = value()

        except (IOError, OSError):
            with self._treant._read:
                try:
                    mdsdict[self._name]
                except KeyError:
                    raise KeyError(
                            ("Missing '{}' data, and cannot write to "
                             "Treant '{}'".format(self._name,
                                                  self._treant.filepath)))

    def _activate(self):
        """Make the universe and attach it.

        """
        if not self.topology:
            self._treant._universe = None
            return

        uh = Universehound(self)
        paths = uh.fetch()
        topology = paths['top']
        trajectory = paths['traj']

        if not trajectory:
            self._treant._universe = Universe(topology, **self.kwargs)
        else:
            self._treant._universe = Universe(topology, *trajectory,
                                              **self.kwargs)

        self._apply_resnums()

        # update the universe definition; will automatically build current
        # path variants for each file
        # if read-only, move on
        try:
            self._set_topology(topology)
            self._set_trajectory(trajectory)
        except OSError:
            warnings.warn(
                "Cannot update paths for universe; "
                " state file is read-only.")

    def reload(self):
        """Re-load the universe from its stored definition.

        """
        self._activate()

    @property
    def topology(self):
        """The topology file for this Sim's universe.

        To change the topology file for this Sim's universe, set the path to
        the file. Setting to ``None`` will disable the Sim's universe,
        but the trajectory paths will be retained.

        """
        with self._treant._read:
            mdsdict = self._treant._state['mdsynthesis']
            topstate = mdsdict['universedef']['topology']

            if not topstate:
                return None
            else:
                return topstate['abspath']

    @topology.setter
    def topology(self, path):
        if isinstance(path, string_types):
            path = path
        elif isinstance(path, Leaf):
            path = path.abspath
        elif path is None:
            pass
        else:
            raise TypeError("Path to topology must be a string or Leaf")

        self._set_topology(path)

        # reset universe, if present
        if self._treant._universe:
            self._activate()

    def _set_topology(self, path):
        with self._treant._write:
            mdsdict = self._treant._state['mdsynthesis']
            topstate = mdsdict['universedef']['topology']

            if path is None:
                mdsdict['universedef']['topology'] = dict()
            else:
                topstate['abspath'] = os.path.abspath(path)
                topstate['relpath'] = os.path.relpath(path,
                                                      self._treant.abspath)

    @property
    def trajectory(self):
        """The trajectory file for this Sim's universe.

        To change the trajectory file(s) for this Sim's Universe, set the path
        to the file. A single path will use a single trajectory file, while a
        list or tuple of paths will use each trajectory file in order for
        building the universe. ``None`` indicates that the Universe will not
        load a trajectory file at all (but the topology may have coordinates).

        """
        with self._treant._read:
            mdsdict = self._treant._state['mdsynthesis']
            traj = mdsdict['universedef']['trajectory']
            if not traj:
                return None
            elif len(traj) == 1:
                return traj[0][0]
            else:
                return tuple([t[0] for t in traj])

    @trajectory.setter
    def trajectory(self, path):
        if isinstance(path, string_types):
            trajs = [path]
        elif isinstance(path, Leaf):
            trajs = [path.abspath]
        elif isinstance(path, (list, tuple)):
            trajs = list(path)
        elif path is None:
            trajs = []
        else:
            raise TypeError("Path to topology must be a string, Leaf, or"
                            " list/tuple of paths.")

        self._set_trajectory(trajs)

        # reset universe, if present
        if self._treant._universe:
            self._activate()

    def _set_trajectory(self, trajs):
        with self._treant._write:
            mdsdict = self._treant._state['mdsynthesis']
            mdsdict['universedef']['trajectory'] = []
            trajstate = mdsdict['universedef']['trajectory']

            for traj in trajs:
                trajstate.append(
                        [os.path.abspath(traj),
                         os.path.relpath(traj, self._treant.abspath)])

    def _apply_resnums(self):
        """Apply resnum definition to universe.

        """
        with self._treant._read:
            simdict = self._treant._state['mdsynthesis']['universedef']
            try:
                resnums = simdict['resnums']
            except KeyError:
                resnums = None

        if resnums:
            self._treant._universe.residues.set_resnums(np.array(resnums))

    @deprecate(message="resnum storage is deprecated")
    def _set_resnums(self, resnums):
        """Define resnums for the universe.

        Resnums are useful for referring to residues by their canonical resid,
        for instance that stored in the PDB. By giving a resnum definition
        for the universe, this definition will be applied to the universe.

        Will overwrite existing resnum definition if it exists.

        Parameters
        ----------
        resnums : array_like
                array or list giving the resnum for each residue in the
                topology, in atom index order; giving ``None`` will delete
                resnum definition
        """
        with self._treant._write:
            simdict = self._treant._state['mdsynthesis']['universedef']
            if resnums is None:
                simdict['resnums'] = None
            else:
                simdict['resnums'] = list(resnums)

            if self._treant._universe:
                self._apply_resnums()

    def _define(self, pathtype='abs'):
        """Get the stored path to the topology and trajectory used for the
        universe.

        .. note:: Does no checking as to whether these paths are valid. To
                  check this, try using the universe.

        Parameters
        ----------
        pathtype : {'abs', 'rel'}
            type of path to return; 'abs' gives an absolute path, 'rel' gives a
            path relative to the Sim's state file

        Returns
        -------
        topology : str
            path to the topology file
        trajectory : list
            list of paths to trajectory files

        """
        with self._treant._read:
            mdsdict = self._treant._state['mdsynthesis']
            top = mdsdict['universedef']['topology']
            outtop = {'abs': top['abspath'],
                      'rel': top['relpath']}

            trajs = mdsdict['universedef']['trajectory']
            outtraj = {key: [] for key in self._filepaths}

        for traj in trajs:
            for i, key in enumerate(self._filepaths):
                outtraj[key].append(traj[i])

        return outtop[pathtype], outtraj[pathtype]

    @property
    def kwargs(self):
        """The keyword arguments applied to the Sim's universe when building
        it.

        Set these with a dictionary of keywords and values to change them.
        Keywords must be strings and values must be strings, ints, floats,
        bools, or ``None``.

        """
        with self._treant._read:
            mdsdict = self._treant._state['mdsynthesis']
            return mdsdict['universedef']['kwargs']

    @kwargs.setter
    def kwargs(self, kwargs):
        if kwargs is None:
            pass
        elif isinstance(kwargs, dict):
            # check that values are serializable
            for key, value in kwargs.items():
                if not (isinstance(value, (string_types, bool, int, float)) or
                        value is None):
                    raise ValueError("Cannot store keyword '{}' for Universe; "
                                     "value must be a string, bool, int, "
                                     "float, or ``None``, "
                                     "not '{}'".format(key, type(value)))
        else:
            raise TypeError("Must be a dictionary or ``None``")

        with self._treant._write:
            simdict = self._treant._state['mdsynthesis']['universedef']
            simdict['kwargs'] = kwargs


class AtomSelections(Limb):
    """Stored atom selections for the universe.

    Useful atom selections can be stored for the universe and recalled later.

    """
    _name = 'atomselections'

    def __init__(self, treant):
        super(AtomSelections, self).__init__(treant)

        # init state if for selections not already there;
        # if read-only, check that it is there,
        # and raise exception if it is not
        try:
            with self._treant._write:
                try:
                    self._treant._state['mdsynthesis'][self._name]
                except KeyError:
                    self._treant._state['mdsynthesis'][self._name] = dict()
        except (IOError, OSError):
            with self._treant._read:
                try:
                    self._treant._state['mdsynthesis'][self._name]
                except KeyError:
                    raise KeyError(
                            ("Missing '{}' data, and cannot write to "
                             "Treant '{}'".format(self._name,
                                                  self._treant.filepath)))

    def __repr__(self):
        return "<AtomSelections({})>".format(
                {x: self.get(x) for x in self.keys()})

    def __getitem__(self, handle):
        """Get selection for given handle.

        Parameters
        ----------
        handle : str
            Name of selection to return.

        Returns
        -------
        selection : str, list, array_like
            The named selection definition.

        """
        return self.get(handle)

    def __setitem__(self, handle, selection):
        """Selection for the given handle.

        """
        if isinstance(selection, (string_types, np.ndarray)):
            selection = [selection]
        self.add(handle, *selection)

    def __iter__(self):
        return self.keys().__iter__()

    def __delitem__(self, handle):
        """Remove stored selection for given handle and the active universe.

        """
        self.remove(handle)

    def add(self, handle, *selection):
        """Add an atom selection for the attached universe.

        AtomGroups are needed to obtain useful information from raw coordinate
        data. It is useful to store AtomGroup selections for later use, since
        they can be complex and atom order may matter.

        If a selection with the given `handle` already exists, it is replaced.

        Parameters
        ----------
        handle : str
            Name to use for the selection.
        selection : str, AtomGroup
            Selection string or AtomGroup indices; multiple selections may be
            given and their order will be preserved, which is useful for e.g.
            structural alignments.

        """
        if len(selection) == 1:
            sel = selection[0]
            if isinstance(sel, np.ndarray):
                outsel = sel.tolist()
            elif isinstance(sel, string_types):
                    outsel = sel
        else:
            outsel = list()
            for sel in selection:
                if isinstance(sel, np.ndarray):
                    outsel.append(sel.tolist())
                elif isinstance(sel, string_types):
                    outsel.append(sel)
                else:
                    raise ValueError("Selections must be strings, arrays of "
                                     "atom indices, or tuples/lists of these.")

        with self._treant._write:
            seldict = self._treant._state['mdsynthesis']['atomselections']
            seldict[handle] = outsel

    def remove(self, *handle):
        """Remove an atom selection for the universe.

        If named selection doesn't exist, :exc:`KeyError` raised.

        Parameters
        ----------
        handle : str
            Name of selection(s) to remove.

        """
        with self._treant._write:
            seldict = self._treant._state['mdsynthesis']['atomselections']
            for item in handle:
                try:
                    del seldict[item]
                except KeyError:
                    raise KeyError("No such selection '{}'".format(item))

    def keys(self):
        """Return a list of all selection handles.

        """
        with self._treant._read:
            seldict = self._treant._state['mdsynthesis']['atomselections']
            return seldict.keys()

    def create(self, handle):
        """Generate AtomGroup from universe from the given named selection.

        If named selection doesn't exist, :exc:`KeyError` raised.

        Parameters
        ----------
        handle : str
            Name of selection to return as an AtomGroup.

        Returns
        -------
        AtomGroup
            The named selection as an AtomGroup of the universe.

        """
        sel = self.get(handle)

        # Selections might be either
        # - a single string
        # - a single list of indices
        # - a list of strings
        # - a list of indices

        if isinstance(sel, string_types):
            # if we have a single string
            ag = self._treant.universe.select_atoms(sel)
        elif all([isinstance(i, int) for i in sel]):
            # if we have a single array_like of indices
            ag = self._treant.universe.atoms[sel]
        else:
            ag = None
            for item in sel:
                if isinstance(item, string_types):
                    if ag:
                        ag += self._treant.universe.select_atoms(item)
                    else:
                        ag = self._treant.universe.select_atoms(item)
                else:
                    if ag:
                        ag += self._treant.universe.atoms[item]
                    else:
                        ag = self._treant.universe.atoms[item]

        return ag

    def get(self, handle):
        """Get selection definition for given handle.

        If named selection doesn't exist, :exc:`KeyError` raised.

        Parameters
        ----------
        handle : str
            Name of selection to get definition of.

        Returns
        -------
        definition : list
            list of strings defining the atom selection

        """
        with self._treant._read:
            seldict = self._treant._state['mdsynthesis']['atomselections']

        try:
            seldef = seldict[handle]
        except KeyError:
            raise KeyError("No such selection '{}'".format(handle))

        if isinstance(seldef, string_types):
            # if we have a single string
            out = seldef
        elif all([isinstance(i, int) for i in seldef]):
            # if we have a single list of indices
            out = np.array(seldef)
        else:
            out = []
            for item in seldef:
                if isinstance(item, string_types):
                    out.append(item)
                else:
                    out.append(np.array(item))

            out = tuple(out)

        return out



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# mdsynthesis

"""
MDSynthesis --- a persistence engine for molecular dynamics data
================================================================
"""
# Bring some often used objects into the current namespace
from datreant.core import Treant, Group, Bundle, Tree, Leaf, View
from datreant.core import discover

from .treants import Sim
from . import attach

__all__ = ['Sim', 'Group', 'Bundle']
__version__ = "0.6.0"  # NOTE: keep in sync with RELEASE in setup.py



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# datreant

"""
mdsynthesis.backends --- low-level state file backends
======================================================

"""



"""
Basic Treant objects: the organizational units for :mod:`mdsynthesis`.

"""
import warnings
import os
from six import string_types

from MDAnalysis import Universe

from datreant.core import Treant, Leaf
from . import limbs
from .backends import statefiles


class Sim(Treant):
    """The Sim object is an interface to data for a single simulation.

    `sim` should be a base directory of a new or existing Sim. An existing
    Sim will be regenerated if a state file is found.  If no state file is
    found, a new Sim will be created.

    A Tree object may also be used in the same way as a directory string.

    If multiple Treant state files are in the given directory, a
    :exc:`MultipleTreantsError` will be raised; specify the full path to the
    desired state file to regenerate the desired Treant in this case. It is
    generally better to avoid having multiple state files in the same
    directory.

    Use the `new` keyword to force generation of a new Sim at the given path.

    Parameters
    ----------
    sim : str or Tree
        Base directory of a new or existing Sim; will regenerate a Sim if a
        state file is found, but will genereate a new one otherwise; may also
        be a Tree object
    new : bool
        Generate a new Sim even if one already exists at the given location
    categories : dict
        dictionary with user-defined keys and values; used to give Sims
        distinguishing characteristics
    tags : list
        list with user-defined values; like categories, but useful for adding
        many distinguishing descriptors
    """
    _treanttype = 'Sim'
    _backendclass = statefiles.SimFile

    def __init__(self, sim, new=False, categories=None, tags=None):
        super(Sim, self).__init__(sim,
                                  new=new,
                                  categories=categories,
                                  tags=tags)

        self._universedef = None
        self._atomselections = None
        self._universe = None     # universe 'dock'

    def __repr__(self):
        return "<{}: '{}'>".format(self._treanttype, self.name)

    @property
    def universe(self):
        """The universe of the Sim.

        Universes are interfaces to raw simulation data, with stored selections
        for this universe directly available via ``Sim.selections``.

        Setting this to a :class:`MDAnalysis.Universe` will set that as the
        universe definition for this Sim. Setting to ``None`` will remove
        the universe definition entirely.

        """
        # TODO: include check for changes to universe definition, not just
        # definition absence
        if self._universe:
            return self._universe
        else:
            self.universedef._activate()
            return self._universe

    @universe.setter
    def universe(self, universe):
        if universe is None:
            self.universedef._set_topology(None)
            self.universedef._set_trajectory([])
            self.universedef.kwargs = None
            self.universedef.activate()

        elif not isinstance(universe, Universe):
            raise TypeError("Cannot set to {}; must be Universe".format(
                                type(universe)))
        else:
            self.universedef._set_topology(universe.filename)
            try:  # ChainReader?
                traj = universe.trajectory.filenames
            except AttributeError:
                try:  # Reader?
                    traj = [universe.trajectory.filename]
                except AttributeError:  # Only topology
                    traj = []

            self.universedef._set_trajectory(traj)

            # try and store keyword arguments
            try:
                self.universedef.kwargs = universe.kwargs
            except AttributeError:
                warnings.warn("Universe did not keep keyword arguments; "
                              "cannot store keyword arguments for Universe.")

            # finally, just use this instance
            self._universe = universe

    @universe.deleter
    def universe(self):
        self.universedef._set_topology(None)
        self.universedef._set_trajectory([])
        self.universedef.kwargs = None
        self.universedef.activate()

    @property
    def universedef(self):
        """The universe definition for this Sim.

        """
        # attach universe if not attached, and only give results if a
        # universe is present thereafter
        if not self._universedef:
            self._universedef = limbs.UniverseDefinition(self)

        return self._universedef

    @property
    def atomselections(self):
        """Stored atom selections for the universe.

        Useful atom selections can be stored for the universe and
        recalled later. Selections are stored separately for each defined
        universe, since the same selection may require a different selection
        string for different universes.

        """
        # attach universe if not attached, and only give results if a
        # universe is present thereafter
        if self.universe:
            if not self._atomselections:
                self._atomselections = limbs.AtomSelections(self)

            return self._atomselections



#! /usr/bin/python
"""Setuptools-based setup script for MDSynthesis.

For a basic installation just type the command::

  python setup.py install

"""

from setuptools import setup, find_packages

setup(name='mdsynthesis',
      version='0.6.0',
      description='a persistence engine for molecular dynamics data',
      author='David Dotson', 
      author_email='dotsdl@gmail.com',
      url='http://mdsynthesis.readthedocs.org/',
      classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development :: Libraries :: Python Modules',
        ],
      packages=find_packages('src'),
      package_dir={'': 'src'},
      license='GPL 2',
      long_description=open('README.rst').read(),
      install_requires=[
                'datreant.core>=0.6.0',
                'datreant.data>=0.6.0',
                'MDAnalysis>=0.14.0',
                ],
     )
